/*
 * Copyright 2018 iserge.
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

package org.cesiumjs.cs.datasources.graphics;

import jsinterop.annotations.JsConstructor;
import jsinterop.annotations.JsOptional;
import jsinterop.annotations.JsProperty;
import jsinterop.annotations.JsType;
import org.cesiumjs.cs.core.Color;
import org.cesiumjs.cs.core.Event;
import org.cesiumjs.cs.core.Math;
import org.cesiumjs.cs.datasources.graphics.options.PolylineVolumeGraphicsOptions;
import org.cesiumjs.cs.datasources.properties.MaterialProperty;
import org.cesiumjs.cs.datasources.properties.Property;
import org.cesiumjs.cs.scene.enums.ShadowMode;

/**
 * @author Serge Silaev aka iSergio
 */
@JsType(isNative = true, namespace = "Cesium", name = "PolylineVolumeGraphics")
public class PolylineVolumeGraphics {
    /**
     * A Property specifying the array of Cartesian3 positions which define the line
     * strip.
     */
    @JsProperty
    public Property positions;
    /**
     * A Property specifying the array of Cartesian2 positions which define the
     * shape to be extruded.
     */
    @JsProperty
    public Property shape;
    /**
     * A CornerType Property specifying the style of the corners. Default:
     * CornerType#ROUNDED
     */
    @JsProperty
    public Property cornerType;
    /**
     * A boolean Property specifying the visibility of the volume. Default: true
     */
    @JsProperty
    public Property show;
    /**
     * A boolean Property specifying whether the volume is filled with the provided
     * material. Default: true
     */
    @JsProperty
    public Property fill;
    /**
     * A Property specifying the material used to fill the volume. Default:
     * {@link org.cesiumjs.cs.core.Color#WHITE()}
     */
    @JsProperty
    public MaterialProperty material;
    /**
     * A boolean Property specifying whether the volume is outlined. Default: false
     */
    @JsProperty
    public Property outline;
    /**
     * A Property specifying the Color of the outline. Default:
     * {@link Color#BLACK()}
     */
    @JsProperty
    public Property outlineColor;
    /**
     * A numeric Property specifying the width of the outline. Default: 1.0
     */
    @JsProperty
    public Property outlineWidth;
    /**
     * A numeric Property specifying the angular distance between each latitude and
     * longitude point. Default: {@link Math#RADIANS_PER_DEGREE()}
     */
    @JsProperty
    public Property granularity;
    /**
     * An enum Property specifying whether the volume casts or receives shadows from
     * each light source. Default: {@link ShadowMode#DISABLED()}
     */
    @JsProperty
    public Property shadows;
    /**
     * A Property specifying at what distance from the camera that this volume will
     * be displayed.
     */
    @JsProperty
    public Property distanceDisplayCondition;

    /**
     * Describes a polyline volume defined as a line strip and corresponding two
     * dimensional shape which is extruded along it. The resulting volume conforms
     * to the curvature of the globe.
     */
    @JsConstructor
    public PolylineVolumeGraphics() {
    }

    /**
     * Describes a polyline volume defined as a line strip and corresponding two
     * dimensional shape which is extruded along it. The resulting volume conforms
     * to the curvature of the globe.
     *
     * @param options Options
     */
    @JsConstructor
    public PolylineVolumeGraphics(PolylineVolumeGraphicsOptions options) {
    }

    /**
     * Gets the event that is raised whenever a property or sub-property is changed
     * or modified.
     */
    @JsProperty(name = "definitionChanged")
    public native Event definitionChanged();

    /**
     * Duplicates this instance.
     *
     * @param result The object onto which to store the result.
     * @return The modified result parameter or a new instance if one was not
     * provided.
     */
    public native PolylineVolumeGraphics clone(@JsOptional PolylineVolumeGraphics result);

    /**
     * Assigns each unassigned property on this object to the value of the same
     * property on the provided source object.
     *
     * @param source The object to be merged into this object.
     */
    public native void merge(PolylineVolumeGraphics source);
}
