/*
 * Copyright 2017 iserge.
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

package org.cleanlogic.cesiumjs4gwt.showcase.examples;

import com.google.gwt.core.client.GWT;
import com.google.gwt.user.client.ui.HTML;
import org.cesiumjs.cs.Cesium;
import org.cesiumjs.cs.collections.TimeIntervalCollection;
import org.cesiumjs.cs.core.Math;
import org.cesiumjs.cs.core.*;
import org.cesiumjs.cs.core.enums.ClockRange;
import org.cesiumjs.cs.core.interpolation.HermitePolynomialApproximation;
import org.cesiumjs.cs.core.options.TimeIntervalOptions;
import org.cesiumjs.cs.datasources.Entity;
import org.cesiumjs.cs.datasources.graphics.ModelGraphics;
import org.cesiumjs.cs.datasources.graphics.options.ModelGraphicsOptions;
import org.cesiumjs.cs.datasources.options.EntityOptions;
import org.cesiumjs.cs.datasources.properties.ConstantProperty;
import org.cesiumjs.cs.datasources.properties.SampledPositionProperty;
import org.cesiumjs.cs.datasources.properties.options.SampledPropertyInterpolationOptions;
import org.cesiumjs.cs.js.JsDate;
import org.cesiumjs.cs.scene.Scene;
import org.cesiumjs.cs.widgets.ViewerPanel;
import org.cesiumjs.cs.widgets.options.ViewerOptions;
import org.cleanlogic.cesiumjs4gwt.showcase.basic.AbstractExample;
import org.cleanlogic.cesiumjs4gwt.showcase.components.store.ShowcaseExampleStore;

import javax.inject.Inject;

/**
 * @author Serge Silaev aka iSergio
 */
public class Cardboard extends AbstractExample {
    private JulianDate start;
    private JulianDate stop;
    private Entity entity;

    @Inject
    public Cardboard(ShowcaseExampleStore store) {
        super("Cardboard", "Configure viewer to add a button enabling look at a mobile device with cardboard",
                new String[]{"Showcase", "Cesium", "3d", "Viewer"}, store);
    }

    @Override
    public void buildPanel() {
        ViewerOptions viewerOptions = new ViewerOptions();
        viewerOptions.vrButton = true;
        viewerOptions.terrainProvider = Cesium.createWorldTerrain();

        ViewerPanel csVPanel = new ViewerPanel(viewerOptions);

        csVPanel.getViewer().scene().globe.enableLighting = true;

        csVPanel.getViewer().scene().globe.depthTestAgainstTerrain = true;

        // Follow the path of a plane. See the interpolation Sandcastle example.
        Math.setRandomNumberSeed(3);

        start = JulianDate.fromDate(new JsDate(2015, 2, 25, 16));
        stop = JulianDate.addSeconds(start, 360, new JulianDate());

        csVPanel.getViewer().clock().startTime = start.clone();
        csVPanel.getViewer().clock().stopTime = stop.clone();
        csVPanel.getViewer().clock().currentTime = start.clone();
        csVPanel.getViewer().clock().clockRange = ClockRange.LOOP_STOP();
        csVPanel.getViewer().clock().multiplier = 1.0;

        double lon = -112.110693;
        double lat = 36.0994841;
        double radius = 0.03;

        String modelURI = GWT.getModuleBaseURL() + "SampleData/models/CesiumBalloon/CesiumBalloon.glb";
        EntityOptions entityOptions = new EntityOptions();
        TimeIntervalOptions timeIntervalOptions = new TimeIntervalOptions();
        timeIntervalOptions.start = start;
        timeIntervalOptions.stop = stop;
        entityOptions.availability = new TimeIntervalCollection(
                new TimeInterval[]{new TimeInterval(timeIntervalOptions)});
        entityOptions.position = computeCirclularFlight(lon, lat, radius);
        ModelGraphicsOptions modelGraphicsOptions = new ModelGraphicsOptions();
        modelGraphicsOptions.uri = new ConstantProperty<>(modelURI);
        modelGraphicsOptions.minimumPixelSize = new ConstantProperty<>(64);
        entityOptions.model = new ModelGraphics(modelGraphicsOptions);
        entity = csVPanel.getViewer().entities().add(entityOptions);

        SampledPropertyInterpolationOptions sampledPropertyInterpolationOptions = new SampledPropertyInterpolationOptions();
        sampledPropertyInterpolationOptions.interpolationDegree = 2;
        sampledPropertyInterpolationOptions.interpolationAlgorithm = HermitePolynomialApproximation.instance();
        ((SampledPositionProperty) entity.position).setInterpolationOptions(sampledPropertyInterpolationOptions);

        // Set initial camera position and orientation to be when in the model's
        // reference frame.
        final org.cesiumjs.cs.scene.Camera camera = csVPanel.getViewer().camera;
        camera.position = new Cartesian3(0.25, 0.0, 0.0);
        camera.direction = new Cartesian3(1.0, 0.0, 0.0);
        camera.up = new Cartesian3(0.0, 0.0, 1.0);
        camera.right = new Cartesian3(0.0, -1.0, 0.0);

        csVPanel.getViewer().scene().preRender().addEventListener((Scene.Listener) (scene, time) -> {
            Cartesian3 position = entity.position.getValue(time);
            if (position == null || !Cesium.defined(position)) {
                return;
            }

            Matrix4 transform;
            if (!Cesium.defined(entity.orientation)) {
                transform = Transforms.eastNorthUpToFixedFrame(position);
            } else {
                Quaternion orientation = (Quaternion) entity.orientation.getValue(time);
                if (!Cesium.defined(orientation)) {
                    return;
                }

                transform = Matrix4.fromRotationTranslation(Matrix3.fromQuaternion(orientation), position);
            }

            // Save camera state
            Cartesian3 offset = camera.position.clone();
            Cartesian3 direction = camera.direction.clone();
            Cartesian3 up = camera.up.clone();

            // Set camera to be in model's reference frame.
            camera.lookAtTransform(transform);

            // Reset the camera state to the saved state so it appears fixed in the model's
            // frame.
            offset.clone(camera.position);
            direction.clone(camera.direction);
            up.clone(camera.up);
            Cartesian3.cross(direction, up, camera.right);
        });

        // Add a few more balloons flying with the one the viewer is in.
        int numBalloons = 12;
        for (int i = 0; i < numBalloons; ++i) {
            double balloonRadius = (Math.nextRandomNumber() * 2.0 - 1.0) * 0.01 + radius;
            entityOptions = new EntityOptions();
            timeIntervalOptions = new TimeIntervalOptions();
            timeIntervalOptions.start = start;
            timeIntervalOptions.stop = stop;
            entityOptions.availability = new TimeIntervalCollection(
                    new TimeInterval[]{new TimeInterval(timeIntervalOptions)});
            entityOptions.position = computeCirclularFlight(lon, lat, balloonRadius);
            modelGraphicsOptions = new ModelGraphicsOptions();
            modelGraphicsOptions.uri = new ConstantProperty<>(modelURI);
            modelGraphicsOptions.minimumPixelSize = new ConstantProperty<>(64);
            entityOptions.model = new ModelGraphics(modelGraphicsOptions);
            Entity balloon = csVPanel.getViewer().entities().add(entityOptions);

            sampledPropertyInterpolationOptions = new SampledPropertyInterpolationOptions();
            sampledPropertyInterpolationOptions.interpolationDegree = 2;
            sampledPropertyInterpolationOptions.interpolationAlgorithm = HermitePolynomialApproximation.instance();
            ((SampledPositionProperty) balloon.position).setInterpolationOptions(sampledPropertyInterpolationOptions);
        }

        contentPanel
                .add(new HTML("<p>Configure viewer to add a button enabling look at a mobile device with cardboard.</p>"));
        contentPanel.add(csVPanel);

        initWidget(contentPanel);
    }

    @Override
    public String[] getSourceCodeURLs() {
        String[] sourceCodeURLs = new String[1];
        sourceCodeURLs[0] = GWT.getModuleBaseURL() + "examples/" + "Cardboard.txt";
        return sourceCodeURLs;
    }

    public SampledPositionProperty computeCirclularFlight(double lon, double lat, double radius) {
        SampledPositionProperty property = new SampledPositionProperty();
        double startAngle = Math.nextRandomNumber() * 360.0;
        double endAngle = startAngle + 360.0;

        double increment = (Math.nextRandomNumber() * 2.0 - 1.0) * 10.0 + 45.0;
        for (double i = startAngle; i < endAngle; i += increment) {
            double radians = Math.toRadians(i);
            double timeIncrement = i - startAngle;
            JulianDate time = JulianDate.addSeconds(start, (int) timeIncrement, new JulianDate());
            Cartesian3 position = Cartesian3.fromDegrees(lon + (radius * 1.5 * java.lang.Math.cos(radians)),
                    lat + (radius * java.lang.Math.sin(radians)), Math.nextRandomNumber() * 500 + 1750);
            property.addSample(time, position);
        }
        return property;
    }
}